@@ -1,8 +1,7 @@
 /*
  * entry.c
- * $Id: entry.c 88874 2012-01-14 03:03:52Z cal@macports.org $
  *
- * Copyright (c) 2010-2011 The MacPorts Project
+ * Copyright (c) 2010-2011, 2014, 2017 The MacPorts Project
  * Copyright (c) 2007 Chris Pickel <sfiera@macports.org>
  * All rights reserved.
  *
@@ -31,6 +30,12 @@
 #include <config.h>
 #endif
 
+/* required for asprintf(3) on OS X */
+#define _DARWIN_C_SOURCE
+/* required for asprintf(3) on Linux */
+#define _GNU_SOURCE
+
+#include "portgroup.h"
 #include "entry.h"
 #include "registry.h"
 #include "sql.h"
@@ -39,12 +44,9 @@
 #include <sqlite3.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stdio.h>
 
 /*
- * TODO: possibly, allow reg_entry_search to take different matching strategies
- *       for different keys. I don't know of an application for this feature
- *       yet, so no reason to bother for now.
- *
  * TODO: reg_entry_installed and reg_entry_imaged could benefit from the added
  *       flexibility of -glob and -regexp too. Not a high priority, though.
  *
@@ -184,11 +186,17 @@ reg_entry* reg_entry_open(reg_registry* reg, char* name, char* version,
     int lower_bound = 0;
     char* query;
     if (strlen(epoch) > 0) {
-        query = "SELECT id FROM registry.ports WHERE name=? AND version=? "
-        "AND revision=? AND variants=? AND epoch=?";
+        query = "SELECT id FROM registry.ports "
+#if SQLITE_VERSION_NUMBER >= 3006004
+                "INDEXED BY port_name "
+#endif
+                "WHERE name=? AND version=? AND revision=? AND variants=? AND epoch=?";
     } else {
-        query = "SELECT id FROM registry.ports WHERE name=? AND version=? "
-        "AND revision=? AND variants=? AND epoch!=?";
+        query = "SELECT id FROM registry.ports "
+#if SQLITE_VERSION_NUMBER >= 3006004
+                "INDEXED BY port_name "
+#endif
+                "WHERE name=? AND version=? AND revision=? AND variants=? AND epoch!=?";
     }
     if ((sqlite3_prepare_v2(reg->db, query, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_text(stmt, 1, name, -1, SQLITE_STATIC)
@@ -246,9 +254,11 @@ int reg_entry_delete(reg_entry* entry, reg_error* errPtr) {
     sqlite3_stmt* ports = NULL;
     sqlite3_stmt* files = NULL;
     sqlite3_stmt* dependencies = NULL;
+    sqlite3_stmt* portgroups = NULL;
     char* ports_query = "DELETE FROM registry.ports WHERE id=?";
     char* files_query = "DELETE FROM registry.files WHERE id=?";
     char* dependencies_query = "DELETE FROM registry.dependencies WHERE id=?";
+    char* portgroups_query = "DELETE FROM registry.portgroups WHERE id=?";
     if ((sqlite3_prepare_v2(reg->db, ports_query, -1, &ports, NULL) == SQLITE_OK)
             && (sqlite3_bind_int64(ports, 1, entry->id) == SQLITE_OK)
             && (sqlite3_prepare_v2(reg->db, files_query, -1, &files, NULL)
@@ -256,7 +266,10 @@ int reg_entry_delete(reg_entry* entry, reg_error* errPtr) {
             && (sqlite3_bind_int64(files, 1, entry->id) == SQLITE_OK)
             && (sqlite3_prepare_v2(reg->db, dependencies_query, -1, &dependencies,
                     NULL) == SQLITE_OK)
-            && (sqlite3_bind_int64(dependencies, 1, entry->id) == SQLITE_OK)) {
+            && (sqlite3_bind_int64(dependencies, 1, entry->id) == SQLITE_OK)
+            && (sqlite3_prepare_v2(reg->db, portgroups_query, -1, &portgroups,
+                    NULL) == SQLITE_OK)
+            && (sqlite3_bind_int64(portgroups, 1, entry->id) == SQLITE_OK)) {
         int r;
         do {
             r = sqlite3_step(ports);
@@ -271,7 +284,20 @@ int reg_entry_delete(reg_entry* entry, reg_error* errPtr) {
                                         r = sqlite3_step(dependencies);
                                         switch (r) {
                                             case SQLITE_DONE:
-                                                result = 1;
+                                                do {
+                                                    r = sqlite3_step(portgroups);
+                                                    switch (r) {
+                                                        case SQLITE_DONE:
+                                                            result = 1;
+                                                            break;
+                                                        case SQLITE_BUSY:
+                                                            break;
+                                                        case SQLITE_ERROR:
+                                                            reg_sqlite_error(reg->db,
+                                                                    errPtr, NULL);
+                                                            break;
+                                                    }
+                                                } while (r == SQLITE_BUSY);
                                                 break;
                                             case SQLITE_BUSY:
                                                 break;
@@ -315,6 +341,9 @@ int reg_entry_delete(reg_entry* entry, reg_error* errPtr) {
     if (dependencies) {
         sqlite3_finalize(dependencies);
     }
+    if (portgroups) {
+        sqlite3_finalize(portgroups);
+    }
     return result;
 }
 
@@ -363,23 +392,19 @@ static int reg_all_entries(reg_registry* reg, char* query, int query_len,
  * @param [in] keys      a list of keys to search by
  * @param [in] vals      a list of values to search by, matching keys
  * @param [in] key_count the number of key/value pairs passed
- * @param [in] strategy  strategy to use (one of the `reg_strategy_*` constants)
+ * @param [in] strategies  strategies to use (one of the `reg_strategy_*` constants)
  * @param [out] entries  a list of matching entries
  * @param [out] errPtr   on error, a description of the error that occurred
  * @return               the number of entries if success; false if failure
  */
 int reg_entry_search(reg_registry* reg, char** keys, char** vals, int key_count,
-        int strategy, reg_entry*** entries, reg_error* errPtr) {
+        int *strategies, reg_entry*** entries, reg_error* errPtr) {
     int i;
     char* kwd = " WHERE ";
     char* query;
     size_t query_len, query_space;
     int result;
-    /* get the strategy */
-    char* op = reg_strategy_op(strategy, errPtr);
-    if (op == NULL) {
-        return -1;
-    }
+    char* op;
     /* build the query */
     query = strdup("SELECT id FROM registry.ports");
     if (!query) {
@@ -387,6 +412,11 @@ int reg_entry_search(reg_registry* reg, char** keys, char** vals, int key_count,
     }
     query_len = query_space = strlen(query);
     for (i=0; i<key_count; i++) {
+        /* get the strategy */
+        op = reg_strategy_op(strategies[i], errPtr);
+        if (op == NULL) {
+            return -1;
+        }
         char* cond = sqlite3_mprintf(op, keys[i], vals[i]);
         if (!cond || !reg_strcat(&query, &query_len, &query_space, kwd)
             || !reg_strcat(&query, &query_len, &query_space, cond)) {
@@ -472,7 +502,11 @@ int reg_entry_installed(reg_registry* reg, char* name, reg_entry*** entries,
     if (name == NULL) {
         format = "%s WHERE state='installed'";
     } else {
-        format = "%s WHERE state='installed' AND name='%q'";
+        format = "%s "
+#if SQLITE_VERSION_NUMBER >= 3006004
+                "INDEXED BY port_name "
+#endif
+                "WHERE state='installed' AND name='%q'";
     }
     query = sqlite3_mprintf(format, select, name);
     result = reg_all_entries(reg, query, -1, entries, errPtr);
@@ -486,16 +520,26 @@ int reg_entry_installed(reg_registry* reg, char* name, reg_entry*** entries,
  *
  * @param [in] reg     registry to search in
  * @param [in] path    path of the file to check ownership of
+ * @param [in] cs      false if check should be performed case-insensitive,
+ *                     true otherwise
  * @param [out] entry  the owner, or NULL if no active port owns the file
  * @param [out] errPtr on error, a description of the error that occurred
  * @return             true if success; false if failure
  */
-int reg_entry_owner(reg_registry* reg, char* path, reg_entry** entry,
+int reg_entry_owner(reg_registry* reg, char* path, int cs, reg_entry** entry,
         reg_error* errPtr) {
     int result = 0;
     sqlite3_stmt* stmt = NULL;
-    char* query = "SELECT id FROM registry.files WHERE actual_path=? AND active";
     int lower_bound = 0;
+    char *query = NULL;
+
+    asprintf(&query, "SELECT id FROM registry.files WHERE (actual_path = ? %s) AND active",
+             cs ? "" : "COLLATE NOCASE");
+
+    if (!query) {
+        return 0;
+    }
+
     if ((sqlite3_prepare_v2(reg->db, query, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_text(stmt, 1, path, -1, SQLITE_STATIC)
                 == SQLITE_OK)) {
@@ -524,6 +568,7 @@ int reg_entry_owner(reg_registry* reg, char* path, reg_entry** entry,
     if (stmt) {
         sqlite3_finalize(stmt);
     }
+    free(query);
     return result;
 }
 
@@ -537,12 +582,22 @@ int reg_entry_owner(reg_registry* reg, char* path, reg_entry** entry,
  *
  * @param [in] reg  registry to find file in
  * @param [in] path path of file to get owner of
+ * @param [in] cs   false if check should be performed case-insensitive,
+ *                  true otherwise
  * @return          id of owner, or 0 for none
  */
-sqlite_int64 reg_entry_owner_id(reg_registry* reg, char* path) {
+sqlite_int64 reg_entry_owner_id(reg_registry* reg, char* path, int cs) {
     sqlite3_stmt* stmt = NULL;
     sqlite_int64 result = 0;
-    char* query = "SELECT id FROM registry.files WHERE actual_path=? AND active";
+    char *query = NULL;
+
+    asprintf(&query, "SELECT id FROM registry.files WHERE (actual_path = ? %s) AND active",
+             cs ? "" : "COLLATE NOCASE");
+
+    if (!query) {
+        return 0;
+    }
+
     if ((sqlite3_prepare_v2(reg->db, query, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_text(stmt, 1, path, -1, SQLITE_STATIC)
                 == SQLITE_OK)) {
@@ -557,6 +612,7 @@ sqlite_int64 reg_entry_owner_id(reg_registry* reg, char* path) {
     if (stmt) {
         sqlite3_finalize(stmt);
     }
+    free(query);
     return result;
 }
 
@@ -669,6 +725,73 @@ int reg_entry_propset(reg_entry* entry, char* key, char* value,
     return result;
 }
 
+/**
+ * Associates a portgroup with given port.
+ *
+ * @param [in] entry      the entry to map the portgroup to
+ * @param [in] name       the portgroup name (e.g. "muniversal")
+ * @param [in] version    the portgroup version (e.g. "1.0")
+ * @param [in] sha256     the sha256 hash of the portgroup file
+ * @param [in] size       the size of the portgroup file in bytes
+ * @param [out] errPtr    on error, a description of the error that occurred
+ * @return                true if success; false if failure
+ */
+int reg_entry_addgroup(reg_entry* entry, char* name, char *version,
+        char *sha256, sqlite_int64 size, reg_error* errPtr) {
+    reg_registry* reg = entry->reg;
+    int result = 1;
+    sqlite3_stmt* stmt = NULL;
+    char* insert = "INSERT INTO registry.portgroups (id, name, version, size, sha256) "
+        "VALUES (?, ?, ?, ?, ?)";
+    if ((sqlite3_prepare_v2(reg->db, insert, -1, &stmt, NULL) == SQLITE_OK)
+            && (sqlite3_bind_int64(stmt, 1, entry->id) == SQLITE_OK)
+            && (sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC) == SQLITE_OK)
+            && (sqlite3_bind_text(stmt, 3, version, -1, SQLITE_STATIC) == SQLITE_OK)
+            && (sqlite3_bind_int64(stmt, 4, size) == SQLITE_OK)
+            && (sqlite3_bind_text(stmt, 5, sha256, -1, SQLITE_STATIC) == SQLITE_OK)) {
+        int r;
+        do {
+            r = sqlite3_step(stmt);
+            switch (r) {
+                case SQLITE_DONE:
+                    sqlite3_reset(stmt);
+                    break;
+                case SQLITE_BUSY:
+                    break;
+                default:
+                    reg_sqlite_error(reg->db, errPtr, insert);
+                    result = 0;
+                    break;
+            }
+        } while (r == SQLITE_BUSY);
+    } else {
+        reg_sqlite_error(reg->db, errPtr, insert);
+        result = 0;
+    }
+    if (stmt) {
+        sqlite3_finalize(stmt);
+    }
+    return result;
+}
+
+/**
+ * Gets a list of portgroups that are used by this port.
+ *
+ * @param [in] entry       a port
+ * @param [out] portgroups a list of portgroups used by the given port
+ * @param [out] errPtr     on error, a description of the error that occurred
+ * @return                 true if success; false if failure
+ */
+int reg_entry_getgroups(reg_entry* entry, reg_portgroup*** portgroups, reg_error* errPtr) {
+    reg_registry* reg = entry->reg;
+    char* query = sqlite3_mprintf("SELECT ROWID FROM portgroups "
+            "WHERE id=%lld",
+            entry->id);
+    int result = reg_all_portgroups(reg, query, -1, portgroups, errPtr);
+    sqlite3_free(query);
+    return result;
+}
+
 /**
  * Maps files to the given port in the filemap. The list of files must not
  * contain files that are already mapped to the given port.
@@ -684,8 +807,8 @@ int reg_entry_map(reg_entry* entry, char** files, int file_count,
     reg_registry* reg = entry->reg;
     int result = 1;
     sqlite3_stmt* stmt = NULL;
-    char* insert = "INSERT INTO registry.files (id, path, mtime, active) "
-        "VALUES (?, ?, 0, 0)";
+    char* insert = "INSERT INTO registry.files (id, path, active) "
+        "VALUES (?, ?, 0)";
     if ((sqlite3_prepare_v2(reg->db, insert, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_int64(stmt, 1, entry->id) == SQLITE_OK)) {
         int i;
@@ -737,7 +860,11 @@ int reg_entry_unmap(reg_entry* entry, char** files, int file_count,
     reg_registry* reg = entry->reg;
     int result = 1;
     sqlite3_stmt* stmt = NULL;
-    char* query = "DELETE FROM registry.files WHERE path=? AND id=?";
+    char* query = "DELETE FROM registry.files "
+#if SQLITE_VERSION_NUMBER >= 3006004
+                  "INDEXED BY file_path "
+#endif
+                  "WHERE path=? AND id=?";
     if ((sqlite3_prepare_v2(reg->db, query, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_int64(stmt, 2, entry->id) == SQLITE_OK)) {
         int i;
@@ -938,8 +1065,16 @@ int reg_entry_activate(reg_entry* entry, char** files, char** as_files,
     sqlite3_stmt* update = NULL;
     char* select_query = "SELECT id FROM registry.files WHERE actual_path=? "
         "AND active";
-    char* update_query = "UPDATE registry.files SET actual_path=?, active=1 "
-        "WHERE path=? AND id=?";
+    char* update_query = "UPDATE registry.files "
+#if SQLITE_VERSION_NUMBER >= 3006004
+        /* if the version of SQLite supports it force the usage of the index on
+         * path, rather than the one on id which has a lot less discriminative
+         * power and leads to very slow queries. This is needed for the new
+         * query planner introduced in 3.8.0 which would not use the correct
+         * index automatically. */
+        "INDEXED BY file_path "
+#endif
+        "SET actual_path=?, active=1 WHERE path=? AND id=?";
 
     /* if as_files wasn't specified, activate as the original files */
     if (as_files == NULL) {
@@ -1037,7 +1172,16 @@ int reg_entry_deactivate(reg_entry* entry, char** files, int file_count,
     int result = 1;
     int i;
     sqlite3_stmt* stmt = NULL;
-    char* query = "UPDATE registry.files SET active=0 WHERE actual_path=? AND id=?";
+    char* query = "UPDATE registry.files "
+#if SQLITE_VERSION_NUMBER >= 3006004
+        /* if the version of SQLite supports it force the usage of the index on
+         * path, rather than the one on id which has a lot less discriminative
+         * power and leads to very slow queries. This is needed for the new
+         * query planner introduced in 3.8.0 which would not use the correct
+         * index automatically. */
+        "INDEXED BY file_actual "
+#endif
+        "SET active=0 WHERE actual_path=? AND id=?";
     if ((sqlite3_prepare_v2(reg->db, query, -1, &stmt, NULL) == SQLITE_OK)
             && (sqlite3_bind_int64(stmt, 2, entry->id) == SQLITE_OK)) {
         for (i=0; i<file_count && result; i++) {
@@ -1179,7 +1323,7 @@ int reg_all_open_entries(reg_registry* reg, reg_entry*** entries) {
     int entry_space = 10;
     Tcl_HashEntry* hash;
     Tcl_HashSearch search;
-    *entries = malloc(10*sizeof(void*));
+    *entries = malloc(entry_space * sizeof(reg_entry*));
     if (!*entries) {
         return -1;
     }
